from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import cast, Any, Sequence, SupportsIndex

import sympy as sp

from ..types import PsType, PsNumericType, create_type, UserTypeSpec

from sympy.logic.boolalg import Boolean

from warnings import warn


def is_loop_counter_symbol(symbol):
    from ..defaults import DEFAULTS

    try:
        return DEFAULTS.spatial_counters.index(symbol)
    except ValueError:
        return None


class DynamicType(ABC):
    """Dynamic data type that will be resolved during kernel creation."""


class DynamicNumericType(DynamicType):
    """Use the default numeric type set for the kernel."""

    def __eq__(self, o: Any) -> bool:
        return type(self) is type(o)

    def __hash__(self) -> int:
        return hash(())

    def __str__(self) -> str:
        return "DynamicNumericType"


class DynamicIndexType(DynamicType):
    """Use the default index type set for the kernel.

    This is guaranteed to be an interger type.
    """

    def __eq__(self, o: Any) -> bool:
        return type(self) is type(o)

    def __hash__(self) -> int:
        return hash(())

    def __str__(self) -> str:
        return "DynamicIndexType"


@dataclass
class DynamicNumericArrayType(DynamicType):
    """An array of the default numeric type with a statically known shape."""

    shape: SupportsIndex | Sequence[SupportsIndex]

    def __eq__(self, o: Any) -> bool:
        return type(self) is type(o) and self.shape == o.shape

    def __hash__(self) -> int:
        return hash(self.shape)

    def __str__(self) -> str:
        return f"DynamicNumericArrayType({self.shape})"


class TypeAtom(sp.Atom):
    """Wrapper around a type to disguise it as a SymPy atom."""

    _dtype: PsType | DynamicType

    def __new__(cls, dtype: PsType | DynamicType):
        obj = super().__new__(cls)
        obj._dtype = dtype
        return obj

    def _sympystr(self, *args, **kwargs):
        return str(self._dtype)

    def get(self) -> PsType | DynamicType:
        return self._dtype

    def _hashable_content(self):
        return (self._dtype,)

    def __getnewargs__(self):
        return (self._dtype,)


def assumptions_from_dtype(dtype: PsType | DynamicType):
    """Derives SymPy assumptions from :class:`PsAbstractType`

    Args:
        dtype (PsAbstractType): a pystencils data type
    Returns:
        A dict of SymPy assumptions
    """
    assumptions = dict()

    match dtype:
        case DynamicIndexType():
            assumptions.update({"integer": True, "real": True})
        case DynamicNumericType():
            assumptions.update({"real": True})
        case PsNumericType():
            if dtype.is_int():
                assumptions.update({"integer": True})
            if dtype.is_uint():
                assumptions.update({"negative": False})
            if dtype.is_int() or dtype.is_float():
                assumptions.update({"real": True})

    return assumptions


class TypedSymbol(sp.Symbol):

    _dtype: PsType | DynamicType

    def __new__(cls, *args, **kwds):
        obj = TypedSymbol.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(
        cls, name: str, dtype: UserTypeSpec | DynamicType, **kwargs
    ):  # TODO does not match signature of sp.Symbol???
        # TODO: also Symbol should be allowed  ---> see sympy Variable
        if not isinstance(dtype, DynamicType):
            dtype = create_type(dtype)

        assumptions = assumptions_from_dtype(dtype)
        assumptions.update(kwargs)

        obj = super(TypedSymbol, cls).__xnew__(cls, name, **assumptions)
        obj._dtype = dtype

        return obj

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(sp.core.cacheit(__new_stage2__))

    @property
    def dtype(self) -> PsType | DynamicType:
        #   mypy: ignore
        return self._dtype

    def _hashable_content(self):
        #   mypy: ignore
        return super()._hashable_content(), hash(self._dtype)

    def __getnewargs__(self):
        return self.name, self.dtype

    def __getnewargs_ex__(self):
        return (self.name, self.dtype), self.assumptions0

    @property
    def canonical(self):
        return self

    @property
    def reversed(self):
        return self

    @property
    def headers(self) -> set[str]:
        return self.dtype.required_headers if isinstance(self.dtype, PsType) else set()


class TypeCast(sp.Function):
    """Explicitly cast an expression to a data type."""

    @staticmethod
    def as_numeric(expr):
        return TypeCast(expr, DynamicNumericType())

    @staticmethod
    def as_index(expr):
        return TypeCast(expr, DynamicIndexType())

    @property
    def expr(self) -> sp.Basic:
        return self.args[0]

    @property
    def dtype(self) -> PsType | DynamicType:
        return cast(TypeAtom, self._args[1]).get()

    def __new__(cls, expr: sp.Basic, dtype: UserTypeSpec | DynamicType | TypeAtom):
        tatom: TypeAtom
        match dtype:
            case TypeAtom():
                tatom = dtype
            case DynamicType():
                tatom = TypeAtom(dtype)
            case _:
                tatom = TypeAtom(create_type(dtype))

        return super().__new__(cls, expr, tatom)

    @classmethod
    def eval(cls, expr: sp.Basic, tatom: TypeAtom) -> TypeCast | None:
        dtype = tatom.get()
        if cls is not BoolCast and isinstance(dtype, PsNumericType) and dtype.is_bool():
            return BoolCast(expr, tatom)

        return None

    def _eval_is_integer(self):
        if self.dtype == DynamicIndexType():
            return True
        if isinstance(self.dtype, PsNumericType) and self.dtype.is_int():
            return True

    def _eval_is_real(self):
        if self.dtype == DynamicIndexType() or self.dtype == DynamicNumericType():
            return True
        if isinstance(self.dtype, PsNumericType) and (
            self.dtype.is_float() or self.dtype.is_int()
        ):
            return True

    def _eval_is_nonnegative(self):
        if isinstance(self.dtype, PsNumericType) and self.dtype.is_uint():
            return True


class BoolCast(TypeCast, Boolean):
    pass


tcast = TypeCast


class CastFunc(TypeCast):
    def __new__(cls, *args, **kwargs):
        warn(
            "CastFunc is deprecated and will be removed in pystencils 2.1. "
            "Use `pystencils.tcast` instead.",
            FutureWarning,
        )
        return TypeCast.__new__(cls, *args, **kwargs)
